@@ -18,22 +18,25 @@
  * License along with this library; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  *
- * $Id: main.c,v 1.69 2004/04/22 14:07:14 danchr Exp $
+ * $Id: main.c,v 1.90 2017/07/02 17:10:00 knik Exp $
  */
 
 #ifdef _MSC_VER
-# define HAVE_LIBMP4V2	1
+# define HAVE_LIBMP4V2  1
 #endif
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
 
-#ifdef HAVE_LIBMP4V2
+#ifdef HAVE_EXTERNAL_LIBMP4V2
+# include <mp4v2/mp4v2.h>
+# define HAVE_LIBMP4V2
+#elif defined(HAVE_LIBMP4V2)
 # include <mp4.h>
 #endif
 
-#define DEFAULT_TNS     1
+#define DEFAULT_TNS     0
 
 #ifdef _WIN32
 #include <windows.h>
@@ -42,7 +45,13 @@
 #include <signal.h>
 #endif
 
-#if defined(__unix__) || defined(__APPLE__)
+/* the BSD derivatives don't define __unix__ */
+#if defined(__APPLE__) || defined(__NetBSD__) || defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__bsdi__)
+#define __unix__
+#endif
+
+#ifdef __unix__
+#include <time.h>
 #include <sys/time.h>
 #include <sys/resource.h>
 #include <unistd.h>
@@ -84,7 +93,7 @@ const char *usage =
   "\t\"%s --long-help\" for a description of all options for FAAC.\n"
   "\t\"%s --license\" for the license terms for FAAC.\n\n";
 
-const char *short_help = 
+const char *short_help =
   "Usage: %s [options] infiles ...\n"
   "Options:\n"
   "  -q <quality>\tSet quantizer quality.\n"
@@ -102,6 +111,7 @@ const char *short_help =
   "MP4 specific options:\n"
 #ifdef HAVE_LIBMP4V2
   "  -w\t\tWrap AAC data in MP4 container. (default for *.mp4 and *.m4a)\n"
+  "  -s\t\tOptimize MP4 container layout after encoding\n"
   "  --artist X\tSet artist to X\n"
   "  --writer X\tSet writer to X\n"
   "  --title X\tSet title to X\n"
@@ -126,26 +136,30 @@ const char *short_help =
   "  <http://www.audiocoding.com/wiki/>\n"
   "\n";
 
-const char *long_help = 
+const char *long_help =
   "Usage: %s [options] infiles ...\n"
   "\n"
   "Quality-related options:\n"
-  "  -q <quality>\tSet default variable bitrate (VBR) quantizer quality in percent\n"
-  "\t\t(default: 100, averages at approx. 120 kbps VBR for a normal \n"
+  "  -q <quality>\tSet default variable bitrate (VBR) quantizer quality in percent.\n"
+  "\t\t(default: 100, averages at approx. 120 kbps VBR for a normal\n"
   "\t\tstereo input file with 16 bit and 44.1 kHz sample rate; max.\n"
-  "\t\tvalue 500, min. 10).Set quantizer quality.\n"
+  "\t\tvalue 500, min. 10).\n"
   "  -b <bitrate>\tSet average bitrate (ABR) to approximately <bitrate> kbps.\n"
+"\t\t(max. value 152 kbps/stereo with a 16 kHz cutoff, can be raised\n"
+  "\t\twith a higher -c setting).\n"
   "  -c <freq>\tSet the bandwidth in Hz (default: automatic, i.e. adapts\n"
   "\t\tmaximum value to input sample rate).\n"
   "\n"
   "Input/output options:\n"
-  "  - <stdin>\tIf you simply use a hyphen/minus sign instead of an input\n"
-  "\t\tfile name, FAAC can encode directly from stdin, thus enabling\n"
-  "\t\tpiping within other applications like foobar2000 (see example\n"
-  "\t\tbelow). The same works for stdout as well, so FAAC can pipe its\n"
-  "\t\toutput to other apps such as mp4live (streaming live AAC\n"
-  "\t\tcontent).\n"
+  "  -\t\t<stdin/stdout>: If you simply use a hyphen/minus sign instead\n"
+  "\t\tof an input file name, FAAC can encode directly from stdin,\n"
+  "\t\tthus enabling piping from other applications and utilities. The\n"
+  "\t\tsame works for stdout as well, so FAAC can pipe its output to\n"
+  "\t\tother apps such as a server.\n"
   "  -o X\t\tSet output file to X (only for one input file)\n"
+  "\t\tonly for one input file; you can use *.aac, *.mp4, *.m4a or\n"
+  "\t\t*.m4b as file extension, and the file format will be set\n"
+  "\t\tautomatically to ADTS or MP4).\n"
   "  -P\t\tRaw PCM input mode (default: off, i.e. expecting a WAV header;\n"
   "\t\tnecessary for input files or bitstreams without a header; using\n"
   "\t\tonly -P assumes the default values for -R, -B and -C in the\n"
@@ -160,41 +174,42 @@ const char *long_help =
   "\t\thave to specify a different position of these two mono channels\n"
   "\t\tin your multichannel input files if they haven't been reordered\n"
   "\t\talready).\n"
-  "  -r\t\taw AAC output mode (i.e. without ADTS headers, needed when\n"
-  "\t\tdirectly using the AAC bitstream in a MP4 container like e.g.\n"
-  "\t\tmp4live does; -w uses this mode automatically).\n"
   "\n"
   "MP4 specific options:\n"
 #ifdef HAVE_LIBMP4V2
-  "  -w\t\tWrap AAC data in MP4 container. (default for *.mp4 and *.m4a)\n"
+  "  -w\t\tWrap AAC data in MP4 container. (default for *.mp4, *.m4a and\n"
+  "\t\t*.m4b)\n"
+  "  -s\t\tOptimize MP4 container layout after encoding.\n"
   "  --artist X\tSet artist to X\n"
-  "  --writer X\tSet writer to X\n"
-  "  --title X\tSet title to X\n"
+  "  --writer X\tSet writer/composer to X\n"
+  "  --title X\tSet title/track name to X\n"
   "  --genre X\tSet genre to X\n"
-  "  --album X\tSet album to X\n"
-  "  --compilation\tSet compilation\n"
+  "  --album X\tSet album/performer to X\n"
+  "  --compilation\tMark as compilation\n"
   "  --track X\tSet track to X (number/total)\n"
   "  --disc X\tSet disc to X (number/total)\n"
   "  --year X\tSet year to X\n"
   "  --cover-art X\tRead cover art from file X\n"
-  "\t\tSupported image formats are gif, jpg, and png.\n"
+  "\t\tSupported image formats are GIF, JPEG, and PNG.\n"
   "  --comment X\tSet comment to X\n"
 #else
   "  MP4 support unavailable.\n"
 #endif
   "\n"
-  "Expert options:\n"
+  "Expert options, only for testing purposes:\n"
 #if !DEFAULT_TNS
-  "  --tns  \tEnable TNS, temporal noise shaping, coding.\n"
+  "  --tns  \tEnable coding of TNS, temporal noise shaping.\n"
 #else
-  "  --no-tns\tDisable TNS, temporal noise shaping, coding.\n"
+  "  --no-tns\tDisable coding of TNS, temporal noise shaping.\n"
 #endif
   "  --no-midside\tDon\'t use mid/side coding.\n"
-  "  --mpeg-vers X\tAAC MPEG version, X can be 2 or 4.\n"
+  "  --mpeg-vers X\tForce AAC MPEG version, X can be 2 or 4\n"
   "  --obj-type X\tAAC object type. (LC (Low Complexity, default), Main or LTP\n"
   "\t\t(Long Term Prediction)\n"
-  "  --shortctl X\tEnforce block type (default: both; 1 = short only; 2 = long\n"
-  "\t\tonly).\n"
+  "  --shortctl X\tEnforce block type (0 = both (default); 1 = no short; 2 = no\n"
+  "\t\tlong).\n"
+  "  -r\t\tGenerate raw AAC bitstream (i.e. without any headers).\n"
+  "\t\tNot advised!!!, RAW AAC files are practically useless!!!\n"
   "\n"
   "Documentation:\n"
   "  --license\tShow the FAAC license.\n"
@@ -320,8 +335,8 @@ void signal_handler(int signal) {
 static int check_image_header(const char *buf) {
   if (!strncmp(buf, "\x89\x50\x4E\x47\x0D\x0A\x1A\x0A", 8))
     return 1; /* PNG */
-  else if (!strncmp(buf, "\xFF\xD8\xFF\xE0", 4) &&
-	   !strncmp(buf + 6, "JFIF\0", 5))
+  else if (!strncmp(buf, "\xFF\xD8\xFF\xE0", 4) ||
+       !strncmp(buf, "\xFF\xD8\xFF\xE1", 4))
     return 1; /* JPEG */
   else if (!strncmp(buf, "GIF87a", 6) || !strncmp(buf, "GIF89a", 6))
     return 1; /* GIF */
@@ -404,6 +419,7 @@ int main(int argc, char *argv[])
     char *audioFileName = NULL;
     char *aacFileName = NULL;
     char *aacFileExt = NULL;
+    int aacFileNameGiven = 0;
 
     float *pcmbuf;
     int *chanmap = NULL;
@@ -422,20 +438,21 @@ int main(int argc, char *argv[])
     FILE *outfile = NULL;
 
 #ifdef HAVE_LIBMP4V2
+    int optimizeFlag = 0;
     MP4FileHandle MP4hFile = MP4_INVALID_FILE_HANDLE;
     MP4TrackId MP4track = 0;
-    int ntracks = 0, trackno = 0;
-    int ndiscs = 0, discno = 0;
+    unsigned int ntracks = 0, trackno = 0;
+    unsigned int ndiscs = 0, discno = 0;
     u_int8_t compilation = 0;
     const char *artist = NULL, *title = NULL, *album = NULL, *year = NULL,
       *genre = NULL, *comment = NULL, *writer = NULL;
     u_int8_t *art = NULL;
     u_int64_t artSize = 0;
-    u_int64_t total_samples = 0;
     u_int64_t encoded_samples = 0;
     unsigned int delay_samples;
     unsigned int frameSize;
 #endif
+    u_int64_t total_samples = 0;
     char *faac_id_string;
     char *faac_copyright_string;
 
@@ -461,7 +478,7 @@ int main(int argc, char *argv[])
     while (1) {
         static struct option long_options[] = {
             { "help", 0, 0, 'h'},
-            { "long-help", 0, 0, '?'},
+            { "long-help", 0, 0, 'H'},
             { "raw", 0, 0, 'r'},
             { "no-midside", 0, 0, NO_MIDSIDE_FLAG},
             { "cutoff", 1, 0, 'c'},
@@ -478,6 +495,7 @@ int main(int argc, char *argv[])
             { "license", 0, 0, 'L'},
 #ifdef HAVE_LIBMP4V2
             { "createmp4", 0, 0, 'w'},
+            { "optimize", 0, 0, 's'},
             { "artist", 1, 0, ARTIST_FLAG},
             { "title", 1, 0, TITLE_FLAG},
             { "album", 1, 0, ALBUM_FLAG},
@@ -487,10 +505,10 @@ int main(int argc, char *argv[])
             { "year", 1, 0, YEAR_FLAG},
             { "cover-art", 1, 0, COVER_ART_FLAG},
             { "comment", 1, 0, COMMENT_FLAG},
-	    { "writer", 1, 0, WRITER_FLAG},
-	    { "compilation", 0, 0, COMPILATION_FLAG},
+        { "writer", 1, 0, WRITER_FLAG},
+        { "compilation", 0, 0, COMPILATION_FLAG},
 #endif
-	    { "pcmswapbytes", 0, 0, 'X'},
+        { "pcmswapbytes", 0, 0, 'X'},
             { 0, 0, 0, 0}
         };
         int c = -1;
@@ -498,7 +516,7 @@ int main(int argc, char *argv[])
 
         c = getopt_long(argc, argv, "Hhb:m:o:rnc:q:PR:B:C:I:X"
 #ifdef HAVE_LIBMP4V2
-                        "w"
+                        "ws"
 #endif
             ,long_options, &option_index);
 
@@ -507,19 +525,20 @@ int main(int argc, char *argv[])
 
         if (!c)
         {
-          dieMessage = short_help;
+          dieMessage = usage;
           break;
         }
 
         switch (c) {
-	case 'o':
-	    {
-	        int l = strlen(optarg);
-		aacFileName = malloc(l+1);
-		memcpy(aacFileName, optarg, l);
-		aacFileName[l] = '\0';
-	    }
-	    break;
+    case 'o':
+        {
+            int l = strlen(optarg);
+        aacFileName = malloc(l+1);
+        memcpy(aacFileName, optarg, l);
+        aacFileName[l] = '\0';
+        aacFileNameGiven = 1;
+        }
+        break;
         case 'r': {
             stream = RAW_STREAM;
             break;
@@ -592,72 +611,75 @@ int main(int argc, char *argv[])
         }
 #ifdef HAVE_LIBMP4V2
         case 'w':
-	    container = MP4_CONTAINER;
+        container = MP4_CONTAINER;
             break;
-	case ARTIST_FLAG:
-	    artist = optarg;
-	    break;
-	case WRITER_FLAG:
-	    writer = optarg;
-	    break;
-	case TITLE_FLAG:
-	    title = optarg;
-	    break;
-	case ALBUM_FLAG:
-	    album = optarg;
-	    break;
-	case TRACK_FLAG:
-	    sscanf(optarg, "%i/%i", &trackno, &ntracks);
-	    break;
-	case DISC_FLAG:
-	    sscanf(optarg, "%i/%i", &discno, &ndiscs);
-	    break;
-	case COMPILATION_FLAG:
-	    compilation = 0x1;
-	    break;
-	case GENRE_FLAG:
-	    genre = optarg;
-	    break;
-	case YEAR_FLAG:
-	    year = optarg;
-	    break;
-	case COMMENT_FLAG:
-	    comment = optarg;
-	    break;
-	case COVER_ART_FLAG: {
-	    FILE *artFile = fopen(optarg, "rb");
+        case 's':
+        optimizeFlag = 1;
+            break;
+    case ARTIST_FLAG:
+        artist = optarg;
+        break;
+    case WRITER_FLAG:
+        writer = optarg;
+        break;
+    case TITLE_FLAG:
+        title = optarg;
+        break;
+    case ALBUM_FLAG:
+        album = optarg;
+        break;
+    case TRACK_FLAG:
+        sscanf(optarg, "%d/%d", &trackno, &ntracks);
+        break;
+    case DISC_FLAG:
+        sscanf(optarg, "%d/%d", &discno, &ndiscs);
+        break;
+    case COMPILATION_FLAG:
+        compilation = 0x1;
+        break;
+    case GENRE_FLAG:
+        genre = optarg;
+        break;
+    case YEAR_FLAG:
+        year = optarg;
+        break;
+    case COMMENT_FLAG:
+        comment = optarg;
+        break;
+    case COVER_ART_FLAG: {
+        FILE *artFile = fopen(optarg, "rb");
 
-	    if(artFile) {
-	        u_int64_t r;
+        if(artFile) {
+            u_int64_t r;
 
-	        fseek(artFile, 0, SEEK_END);
-		artSize = ftell(artFile);
+            fseek(artFile, 0, SEEK_END);
+        artSize = ftell(artFile);
 
-		art = malloc(artSize);
+        art = malloc(artSize);
 
-	        fseek(artFile, 0, SEEK_SET);
-		clearerr(artFile);
+            fseek(artFile, 0, SEEK_SET);
+        clearerr(artFile);
 
-		r = fread(art, artSize, 1, artFile);
+        r = fread(art, artSize, 1, artFile);
 
-		if (r != 1) {
-		    dieMessage = "Error reading cover art file!\n";
-		    free(art);
-		    art = NULL;
-		} else if (artSize < 12 || !check_image_header(art)) {
-		    /* the above expression checks the image signature */
-		    dieMessage = "Unsupported cover image file format!\n";
-		    free(art);
-		    art = NULL;
-		}
+        if (r != 1) {
+            dieMessage = "Error reading cover art file!\n";
+            free(art);
+            art = NULL;
+        } else if (artSize < 12 || !check_image_header(art)) {
+            /* the above expression checks the image signature */
+            dieMessage = "Unsupported cover image file format!\n";
+            free(art);
+            art = NULL;
+        }
 
-		fclose(artFile);
-	    } else {
-	        dieMessage = "Error opening cover art file!\n";
-	    }
+        fclose(artFile);
+        } else {
+            dieMessage = "Error opening cover art file!\n";
+        }
 
-	    break;
-	}
+        break;
+    }
 #endif
         case SHORTCTL_FLAG:
             shortctl = atoi(optarg);
@@ -668,7 +690,7 @@ int main(int argc, char *argv[])
         case NO_TNS_FLAG:
             useTns = 0;
             break;
-	case MPEGVERS_FLAG:
+    case MPEGVERS_FLAG:
             mpegVersion = atoi(optarg);
             switch(mpegVersion)
             {
@@ -679,78 +701,81 @@ int main(int argc, char *argv[])
                 mpegVersion = MPEG4;
                 break;
             default:
-	        dieMessage = "Unrecognised MPEG version!\n";
+            dieMessage = "Unrecognised MPEG version!\n";
             }
             break;
-	case OBJTYPE_FLAG:
-	    if (!strcasecmp(optarg, "LC"))
+#if 0
+    case OBJTYPE_FLAG:
+        if (!strcasecmp(optarg, "LC"))
                 objectType = LOW;
-	    else if (!strcasecmp(optarg, "Main"))
-	        objectType = MAIN;
-	    else if (!strcasecmp(optarg, "LTP")) {
-	        mpegVersion = MPEG4;
-		objectType = LTP;
-	    } else
-	        dieMessage = "Unrecognised object type!\n";
-	    break;
+        else if (!strcasecmp(optarg, "Main"))
+            objectType = MAIN;
+        else if (!strcasecmp(optarg, "LTP")) {
+            mpegVersion = MPEG4;
+        objectType = LTP;
+        } else
+            dieMessage = "Unrecognised object type!\n";
+        break;
+#endif
         case 'L':
-	    printf(faac_copyright_string);
-	    dieMessage = license;
-	    break;
-	case 'X':
-	  rawEndian = 0;
-	  break;
-	case '?':
-	  dieMessage = long_help;
-	  break;
-	case 'h':
+        fprintf(stderr, "%s", faac_copyright_string);
+        dieMessage = license;
+        break;
+    case 'X':
+      rawEndian = 0;
+      break;
+    case 'H':
+      dieMessage = long_help;
+      break;
+    case 'h':
           dieMessage = short_help;
-	  break;
+      break;
+    case '?':
         default:
-	  dieMessage = usage;
+      dieMessage = usage;
           break;
         }
     }
 
     /* check that we have at least one non-option arguments */
-    if ((argc - optind) > 1 && aacFileName)
+    if (!dieMessage && (argc - optind) > 1 && aacFileNameGiven)
         dieMessage = "Cannot encode several input files to one output file.\n";
-    else if ((argc - optind) > 1)
-        dieMessage = "Multiple input files not supported yet.\n";
 
-    if ((argc - optind) < 1 || dieMessage)
+    if (argc - optind < 1 || dieMessage)
     {
-      fprintf(stderr, dieMessage ? dieMessage : usage,
-	       progName, progName, progName, progName);
+        fprintf(stderr, dieMessage ? dieMessage : usage,
+           progName, progName, progName, progName);
         return 1;
     }
 
+    while (argc - optind > 0) {
+
     /* get the input file name */
     audioFileName = argv[optind++];
 
     /* generate the output file name, if necessary */
-    if (!aacFileName) {
+    if (!aacFileNameGiven) {
         char *t = strrchr(audioFileName, '.');
-	int l = t ? strlen(audioFileName) - strlen(t) : strlen(audioFileName);
+    int l = t ? strlen(audioFileName) - strlen(t) : strlen(audioFileName);
 
 #ifdef HAVE_LIBMP4V2
-	aacFileExt = container == MP4_CONTAINER ? ".m4a" : ".aac";
+    aacFileExt = container == MP4_CONTAINER ? ".m4a" : ".aac";
 #else
-	aacFileExt = ".aac";
+    aacFileExt = ".aac";
 #endif
 
-	aacFileName = malloc(l+1+4);
-	memcpy(aacFileName, audioFileName, l);
-	memcpy(aacFileName + l, aacFileExt, 4);
-	aacFileName[l+4] = '\0';
+    aacFileName = malloc(l+1+4);
+    memcpy(aacFileName, audioFileName, l);
+    memcpy(aacFileName + l, aacFileExt, 4);
+    aacFileName[l+4] = '\0';
     } else {
         aacFileExt = strrchr(aacFileName, '.');
 
-        if (aacFileExt && (!strcmp(".m4a", aacFileExt) || !strcmp(".mp4", aacFileExt)))
+        if (aacFileExt && (!strcmp(".m4a", aacFileExt) || !strcmp(".m4b", aacFileExt) || !strcmp(".mp4", aacFileExt)))
 #ifndef HAVE_LIBMP4V2
-	    fprintf(stderr, "WARNING: MP4 support unavailable!\n");
+        fprintf(stderr, "WARNING: MP4 support unavailable!\n");
 #else
-	    container = MP4_CONTAINER;
+        container = MP4_CONTAINER;
 #endif
     }
 
@@ -758,14 +783,14 @@ int main(int argc, char *argv[])
     if (rawChans > 0) // use raw input
     {
         infile = wav_open_read(audioFileName, 1);
-	if (infile)
-	{
-	    infile->bigendian = rawEndian;
-	    infile->channels = rawChans;
-	    infile->samplebytes = rawBits / 8;
-	    infile->samplerate = rawRate;
-	    infile->samples /= (infile->channels * infile->samplebytes);
-	}
+    if (infile)
+    {
+        infile->bigendian = rawEndian;
+        infile->channels = rawChans;
+        infile->samplebytes = rawBits / 8;
+        infile->samplerate = rawRate;
+        infile->samples /= (infile->channels * infile->samplebytes);
+    }
     }
     else // header input
         infile = wav_open_read(audioFileName, 0);
@@ -773,7 +798,7 @@ int main(int argc, char *argv[])
     if (infile == NULL)
     {
         fprintf(stderr, "Couldn't open input file %s\n", audioFileName);
-	return 1;
+    return 1;
     }
 
 
@@ -781,20 +806,27 @@ int main(int argc, char *argv[])
     hEncoder = faacEncOpen(infile->samplerate, infile->channels,
         &samplesInput, &maxBytesOutput);
 
+    if (hEncoder == NULL)
+    {
+        fprintf(stderr, "Couldn't open encoder instance for input file %s\n", audioFileName);
+        wav_close(infile);
+        return 1;
+    }
+
 #ifdef HAVE_LIBMP4V2
     if (container != MP4_CONTAINER && (ntracks || trackno || artist ||
-				       title ||  album || year || art ||
-				       genre || comment || discno || ndiscs ||
-				       writer || compilation))
+                       title ||  album || year || art ||
+                       genre || comment || discno || ndiscs ||
+                       writer || compilation))
     {
         fprintf(stderr, "Metadata requires MP4 output!\n");
-	return 1;
+    return 1;
     }
 
     if (container == MP4_CONTAINER)
     {
         mpegVersion = MPEG4;
-	stream = RAW_STREAM;
+    stream = RAW_STREAM;
     }
 
     frameSize = samplesInput/infile->channels;
@@ -848,7 +880,11 @@ int main(int argc, char *argv[])
     if (!faacEncSetConfiguration(hEncoder, myFormat)) {
         fprintf(stderr, "Unsupported output format!\n");
 #ifdef HAVE_LIBMP4V2
+#ifdef MP4_CLOSE_DO_NOT_COMPUTE_BITRATE /* r479 fix */
+        if (container == MP4_CONTAINER) MP4Close(MP4hFile, 0);
+#else
         if (container == MP4_CONTAINER) MP4Close(MP4hFile);
+#endif
 #endif
         return 1;
     }
@@ -858,14 +894,18 @@ int main(int argc, char *argv[])
     if (container == MP4_CONTAINER) {
         unsigned char *ASC = 0;
         unsigned long ASCLength = 0;
-	char *version_string;
+    char *version_string;
 
-#ifdef MP4_CREATE_EXTENSIBLE_FORMAT
-	/* hack to compile against libmp4v2 >= 1.0RC3
-	 * why is there no version identifier in mp4.h? */
+    if (!strcmp(aacFileName, "-"))
+    {
+	fprintf(stderr, "cannot encode MP4 to stdout\n");
+        return 1;
+    }
+
+#ifdef MP4_DETAILS_ERROR /* r453 fix */
         MP4hFile = MP4Create(aacFileName, MP4_DETAILS_ERROR, 0);
 #else
-	MP4hFile = MP4Create(aacFileName, MP4_DETAILS_ERROR, 0, 0);
+        MP4hFile = MP4Create(aacFileName, 0);
 #endif
         if (!MP4_IS_VALID_FILE_HANDLE(MP4hFile)) {
             fprintf(stderr, "Couldn't create output file %s\n", aacFileName);
@@ -877,35 +917,84 @@ int main(int argc, char *argv[])
         MP4SetAudioProfileLevel(MP4hFile, 0x0F);
         faacEncGetDecoderSpecificInfo(hEncoder, &ASC, &ASCLength);
         MP4SetTrackESConfiguration(MP4hFile, MP4track, ASC, ASCLength);
-	free(ASC);
+    free(ASC);
 
-	/* set metadata */
-	version_string = malloc(strlen(faac_id_string) + 6);
-	strcpy(version_string, "FAAC ");
-	strcpy(version_string + 5, faac_id_string);
-	MP4SetMetadataTool(MP4hFile, version_string);
-	free(version_string);
+    /* set metadata */
+#if HAVE_DECL_MP4TAGSALLOC
+    const MP4Tags* tags;
+    tags = MP4TagsAlloc();
+    MP4TagsFetch( tags, MP4hFile );
+#endif
+    version_string = malloc(strlen(faac_id_string) + 6);
+    strcpy(version_string, "FAAC ");
+    strcpy(version_string + 5, faac_id_string);
+#if !HAVE_DECL_MP4TAGSALLOC
+    MP4SetMetadataTool(MP4hFile, version_string);
+#else
+    MP4TagsSetEncodingTool(tags, version_string);
+#endif
+    free(version_string);
 
-	if (artist) MP4SetMetadataArtist(MP4hFile, artist);
-	if (writer) MP4SetMetadataWriter(MP4hFile, writer);
-	if (title) MP4SetMetadataName(MP4hFile, title);
-	if (album) MP4SetMetadataAlbum(MP4hFile, album);
-	if (trackno > 0) MP4SetMetadataTrack(MP4hFile, trackno, ntracks);
-	if (discno > 0) MP4SetMetadataDisk(MP4hFile, discno, ndiscs);
-	if (compilation) MP4SetMetadataCompilation(MP4hFile, compilation);
-	if (year) MP4SetMetadataYear(MP4hFile, year);
-	if (genre) MP4SetMetadataGenre(MP4hFile, genre);
-	if (comment) MP4SetMetadataComment(MP4hFile, comment);
+#if !HAVE_DECL_MP4TAGSALLOC
+    if (artist) MP4SetMetadataArtist(MP4hFile, artist);
+    if (writer) MP4SetMetadataWriter(MP4hFile, writer);
+    if (title) MP4SetMetadataName(MP4hFile, title);
+    if (album) MP4SetMetadataAlbum(MP4hFile, album);
+    if (trackno > 0) MP4SetMetadataTrack(MP4hFile, trackno, ntracks);
+    if (discno > 0) MP4SetMetadataDisk(MP4hFile, discno, ndiscs);
+    if (compilation) MP4SetMetadataCompilation(MP4hFile, compilation);
+    if (year) MP4SetMetadataYear(MP4hFile, year);
+    if (genre) MP4SetMetadataGenre(MP4hFile, genre);
+    if (comment) MP4SetMetadataComment(MP4hFile, comment);
         if (artSize) {
-	    MP4SetMetadataCoverArt(MP4hFile, art, artSize);
-	    free(art);
-	}
+        MP4SetMetadataCoverArt(MP4hFile, art, artSize);
+#else
+    if (artist) MP4TagsSetArtist(tags, artist);
+    if (writer) MP4TagsSetComposer(tags, writer);
+    if (title) MP4TagsSetName(tags, title);
+    if (album) MP4TagsSetAlbum(tags, album);
+    if (trackno > 0) {
+        MP4TagTrack tt;
+        tt.index = trackno;
+        tt.total = ntracks;
+        MP4TagsSetTrack(tags, &tt);
+    }
+    if (discno > 0) {
+        MP4TagDisk td;
+        td.index = discno;
+        td.total = ndiscs;
+        MP4TagsSetDisk(tags, &td);
+    }
+    if (compilation) MP4TagsSetCompilation(tags, compilation);
+    if (year) MP4TagsSetReleaseDate(tags, year);
+    if (genre) MP4TagsSetGenre(tags, genre);
+    if (comment) MP4TagsSetComments(tags, comment);
+    if (artSize) {
+        MP4TagArtwork mp4art;
+        mp4art.data = art;
+        mp4art.size = artSize;
+        mp4art.type = MP4_ART_UNDEFINED; // delegate typing to libmp4v2
+        MP4TagsAddArtwork( tags, &mp4art );
+#endif
+        free(art);
+    }
+#if HAVE_DECL_MP4TAGSALLOC
+    MP4TagsStore( tags, MP4hFile );
+    MP4TagsFree( tags );
+#endif
     }
     else
     {
 #endif
         /* open the aac output file */
-        outfile = fopen(aacFileName, "wb");
+        if (!strcmp(aacFileName, "-"))
+        {
+            outfile = stdout;
+        }
+        else
+        {
+            outfile = fopen(aacFileName, "wb");
+        }
         if (!outfile)
         {
             fprintf(stderr, "Couldn't create output file %s\n", aacFileName);
@@ -920,7 +1009,7 @@ int main(int argc, char *argv[])
     bitRate = myFormat->bitRate;
     if (bitRate)
       fprintf(stderr, "Average bitrate: %d kbps\n",
-	      (bitRate + 500)/1000*infile->channels);
+          (bitRate + 500)/1000*infile->channels);
     fprintf(stderr, "Quantization quality: %ld\n", quantqual);
     fprintf(stderr, "Bandwidth: %d Hz\n", cutOff);
     fprintf(stderr, "Object type: ");
@@ -943,19 +1032,19 @@ int main(int argc, char *argv[])
         fprintf(stderr, " + M/S");
     fprintf(stderr, "\n");
 
-    fprintf(stderr, "File format: ");
+    fprintf(stderr, "Container format: ");
     switch(container)
     {
     case NO_CONTAINER:
       switch(stream)
-	{
-	case RAW_STREAM:
-	  fprintf(stderr, "Headerless AAC (RAW)\n");
-	  break;
-	case ADTS_STREAM:
-	  fprintf(stderr, "MPEG-2 AAC (ADTS)\n");
-	  break;
-	}
+    {
+    case RAW_STREAM:
+      fprintf(stderr, "Headerless AAC (RAW)\n");
+      break;
+    case ADTS_STREAM:
+      fprintf(stderr, "Transport Stream (ADTS)\n");
+      break;
+    }
         break;
 #ifdef HAVE_LIBMP4V2
     case MP4_CONTAINER:
@@ -983,24 +1072,29 @@ int main(int argc, char *argv[])
         fprintf(stderr, "Encoding %s to %s\n", audioFileName, aacFileName);
         if (frames != 0)
             fprintf(stderr, "   frame          | bitrate | elapsed/estim | "
-		    "play/CPU | ETA\n");
+            "play/CPU | ETA\n");
         else
             fprintf(stderr, " frame | elapsed | play/CPU\n");
 
         /* encoding loop */
 #ifdef _WIN32
-	for (;;)
+    for (;;)
 #else
         while (running)
 #endif
         {
             int bytesWritten;
 
-            samplesRead = wav_read_float32(infile, pcmbuf, samplesInput, chanmap);
+            if (total_samples < infile->samples || infile->samples == 0)
+                samplesRead = wav_read_float32(infile, pcmbuf, samplesInput,
+                                               chanmap);
+            else
+                samplesRead = 0;
+
+            if (total_samples + (samplesRead / infile->channels) > infile->samples && infile->samples != 0)
+                samplesRead = (infile->samples - total_samples) * infile->channels;
 
-#ifdef HAVE_LIBMP4V2
             total_samples += samplesRead / infile->channels;
-#endif
 
             /* call the actual encoding routine */
             bytesWritten = faacEncEncode(hEncoder,
@@ -1013,7 +1107,7 @@ int main(int argc, char *argv[])
             {
                 currentFrame++;
                 showcnt--;
-		totalBytesWritten += bytesWritten;
+        totalBytesWritten += bytesWritten;
             }
 
             if ((showcnt <= 0) || !bytesWritten)
@@ -1045,8 +1139,8 @@ int main(int argc, char *argv[])
                         fprintf(stderr,
                             "\r%5d/%-5d (%3d%%)|  %5.1f  | %6.1f/%-6.1f | %7.2fx | %.1f ",
                             currentFrame, frames, currentFrame*100/frames,
-			    ((double)totalBytesWritten * 8.0 / 1000.0) /
-			    ((double)infile->samples / infile->samplerate * currentFrame / frames),
+                ((double)totalBytesWritten * 8.0 / 1000.0) /
+                ((double)infile->samples / infile->samplerate * currentFrame / frames),
                             timeused,
                             timeused * frames / currentFrame,
                             (1024.0 * currentFrame / infile->samplerate) / timeused,
@@ -1104,15 +1198,31 @@ int main(int argc, char *argv[])
 #endif
             }
         }
-        fprintf(stderr, "\n\n");
 
 #ifdef HAVE_LIBMP4V2
         /* clean up */
         if (container == MP4_CONTAINER)
+        {
+#ifdef MP4_CLOSE_DO_NOT_COMPUTE_BITRATE /* r479 fix */
+            MP4Close(MP4hFile, 0);
+#else
             MP4Close(MP4hFile);
-        else
+#endif
+            if (optimizeFlag == 1)
+            {
+                fprintf(stderr, "\n\nMP4 format optimization... ");
+#ifdef MP4_DETAILS_ERROR /* r453 fix */
+                MP4Optimize(aacFileName, NULL, 0);
+#else
+                MP4Optimize(aacFileName, NULL);
+#endif
+                fprintf(stderr, "Done!");
+            }
+        } else
 #endif
             fclose(outfile);
+
+        fprintf(stderr, "\n\n");
     }
 
     faacEncClose(hEncoder);
@@ -1121,12 +1231,86 @@ int main(int argc, char *argv[])
 
     if (pcmbuf) free(pcmbuf);
     if (bitbuf) free(bitbuf);
+    if (aacFileNameGiven) free(aacFileName);
+
+    }
 
     return 0;
 }
 
 /*
 $Log: main.c,v $
+Revision 1.90  2017/07/02 17:10:00  knik
+exit when trying to encode mp4 to stdin
+
+Revision 1.89  2017/07/01 08:52:27  knik
+fixed CVE-2017-9130 (crash with improper .wav input)
+
+Revision 1.88  2015/11/26 14:27:05  knik
+bugfix by Sebastian Wilhelmi: faac exits immediately when encoding raw wav file
+
+Revision 1.87  2015/06/09 10:58:00  knik
+fixed compilation without limbp4v2
+
+Revision 1.86  2012/02/25 17:34:01  knik
+Fix format string security error.
+
+Revision 1.85  2012/02/23 13:26:45  knik
+Support for libmp4v2 r479; Port faac to the iTMF Generic and Tags API.
+Authors: Jaakko Perttilä, Sebastien NOEL, Fabian Greffrath
+
+Revision 1.84  2010/09/27 10:32:17  knik
+Patch by Arthur Yarwood: read correct number of samples from data chunk
+
+Revision 1.83  2009/07/27 18:12:38  menno
+FAAC now able to use external updated libmp4v2
+
+Revision 1.82  2009/01/24 01:10:20  menno
+Made JPEG detection less restrictive
+
+Revision 1.81  2009/01/24 00:15:12  menno
+Added -s option for output of optimized mp4 layout
+
+Revision 1.80  2007/03/19 19:57:40  menno
+Made faacgui buildable again
+Made stdout as output possible (use - as output filename)
+Disabled MAIN and LTP encoding
+Increase version number
+
+Revision 1.79  2006/09/25 19:53:36  menno
+Fixed track and disk number reading from command line
+
+Revision 1.78  2004/12/08 11:07:17  menno
+Make long help work again
+
+Revision 1.77  2004/08/19 15:33:30  menno
+typo
+it's not bad to have this option, but people should be warned that they can get severe playback problems with RAW AAC files (anything other then 44100 will not be decoded properly unless you know that it has that samplerate). Seeking is also not possible on these files.
+
+Revision 1.76  2004/08/19 13:18:44  menno
+Removed stupid comment in help of FAAC. RAW AAC files are USELESS, it seems that already some people encoded their collection using the -r option.
+
+Revision 1.75  2004/08/06 19:33:19  danchr
+TNS is no longer enabled by default (reported by guruboolez)
+documentation fixes in frontend
+default to mp4 for *.m4b as well
+
+Revision 1.74  2004/08/03 00:08:51  danchr
+fix --shortctl documentation
+
+Revision 1.73  2004/08/02 20:53:23  danchr
+*BSD portability fix
+
+Revision 1.72  2004/08/02 20:41:59  danchr
+NetBSD portability fix + fixing metadata bugs w/ sscanf()
+
+Revision 1.71  2004/07/28 08:18:21  danchr
+Darwin portability fixes, should help on Linux too
+
+Revision 1.70  2004/05/03 11:39:05  danchr
+fix documentation bugs (per Hans-Jrgen's suggestions)
+enable (preliminary) multiple output file support
+
 Revision 1.69  2004/04/22 14:07:14  danchr
 set copyright notice to my full name
 
